/*
 *				**** wrp CSV Implementation ****
 *
 * This file contains the implementation of a CSV parser
 * that is based on Kernighan & Pike's "A Practice of Programming"
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "wrpcsv.h"

struct wrpcsv {
	char *line;		/* input chars */
   char *sline;	/* line copy used by split */
   int maxline;	/* size of line[] & sline[] */
   char **field;	/* field pointers */
   int maxfield;	/* size of field[] */
   int numfields;	/* number of fields in use */
  char *fieldsep;	/* field separators */
};

/* global variables */
static int FirstTime = 1;	/* flag for initial call */
static int MaxEnt = 100;	/* default is no more than 100 */
static struct wrpcsv **Csv = NULL;	/* the actual CSV ADTs */
static int *Index;	/* table of keys for CSV IDs */
static char *InLine = NULL;	/* input line for reading from files */
static char *ErrMsg = NULL;	/* most recent error message */

/* default values */
static int maxline = BUFSIZ;	/* max size of input lines */
static int maxfield = 100;	/* max number of fields */
static char *fieldsep = NULL;	/* default field separator chars */

/********************************************************
 * private routines
 *******************************************************/

static int init(void)
{
  int i;

  randomize();
  if (fieldsep == NULL) fieldsep = strdup(",");	/* default delimeters */
  if ((Csv = (malloc(sizeof(struct wrpcsv*) * MaxEnt))) == NULL)
    {
      ErrMsg = "Can't allocate memory for table of Csv";
      MaxEnt = 0;	/* reset max entries */
      return WRPCSV_MEMORY;
    }
  if ((Index = (int *)malloc(sizeof(int) * MaxEnt)) == NULL)
  {
  		ErrMsg = "Can't allocate memory for Index table for CSV";
      MaxEnt = 0; /* reset max entries */
      return WRPCSV_MEMORY;
  }

  for (i=0; i < MaxEnt; i++)
  {
  		Csv[i] = NULL; /* initialize array */
      Index[i] = -1; /* no key yet */
  }

  /* create input buffer */
  if ((InLine = malloc(maxline)) == NULL)
    { /* problems allocating memory */
      ErrMsg = "Can't allocate memory for input buffer";
      return WRPCSV_MEMORY;
    }

  FirstTime = 0;		/* turn off first time flag */

  return 0;
}

static struct wrpcsv *create_csv(void)
{
	struct wrpcsv *csv;

   if ((csv = malloc(sizeof(*csv))) != NULL)
   { /* initialize structure */
   	csv->line = NULL;
      csv->sline = NULL;
      csv->maxline = maxline;	/* default max size of line */
      csv->field = NULL;
      csv->maxfield = maxfield;
      csv->numfields = 0;
      csv->fieldsep = strdup(fieldsep);	/* default delimeters */
   }

   return csv;
}

void free_csv(struct wrpcsv *csv)
{
  int i;

  ErrMsg = "NULL passed to free_csv";
  if (csv == NULL) return; /* bad address */

  if (csv->line != NULL) free(csv->line);
  if (csv->sline != NULL) free(csv->sline);
  if (csv->field != NULL)
    { /* fields present - free them */
      for (i=0; i < csv->maxfield; i++)
	{
	  if (csv->field[i] != NULL) free(csv->field[i]);
	}
      free(csv->field); /* free table also */
    }
  if (csv->fieldsep != NULL) free(csv->fieldsep);
}


static int find_free_csv(void)
{
  int i, k, m;

   for (i=0; i < MaxEnt; i++)
   {
   	if (Csv[i] == NULL)
      {
      	k = rand();	/* generate random number */
         m = k % MaxEnt;	/* drop lower digits */
         Index[i] = k - m + i;	/* generate key for a return value */
      	return i; /* location of free CSV cell */
      }
   }

   /* no free entry available */
   ErrMsg = "No free CSV Entries";
   return -1;	/* table full */
}

struct wrpcsv *get_csv(int key)
{
   struct wrpcsv *csv;
   int id;

   id = key % MaxEnt;
   if (Index[id] != key) csv = NULL;
   else csv = Csv[id]; /* csv entry (or NULL if empty) */

   if (csv == NULL) ErrMsg = "Invalid CSV ID";

   return csv;
}

static int readline(char *buf, FILE *fp)
{
  char *p;

  if (fgets(InLine, maxline, fp) == NULL)
    {
      ErrMsg = "EOF on input file";
      return WRPCSV_EOF;		/* end of file */
    }

  /* find end of line and strip end of line sequence */
  InLine[maxline-1] = '\0';	/* make sure string is terminated */
  for (p=InLine; *p != '\0' && *p != '\n' && *p != '\r'; p++);
  *p = '\0';			/* terminate string */

  return WRPCSV_OK;
}

static char *advquoted(struct wrpcsv *csv, char *p)
{ /* return end of quoted string, stripping extra quotes */
  char *q;			/* target string */

  for (q = p; *p != '\0'; p++, q++)
    { /* copy chars, but skip consecutive quotes */
      if (*p == '"')
	{ /* found a quote - might be end of field */
	  if (*(p+1) == '"')
	    { /* duplicate quote - ignore it */
	      p++; /* skip duplicate quote */
	    }
	  else
	    { /* end of field */
	      p++;		/* point closing quote */
	      break;
	    }
	}
      *q = *p;			/* copy character */
    }
  *q = '\0';			/* terminate string */

  return p;
    }

/*
 * Internal routine to parse a string associated with a CSV
 * The calling routine has already determined that the CSV is valid
 * and the string is non-null
 */
static int parse_csv(struct wrpcsv *csv)
{ /* parse string associated with csv */
  int ret;			/* return code */
  char *p, *q, *t;		/* start and end of string */
  char c;			/* temporary delimeter character */
  size_t v;			/* amount to allocate */
  int i;

  csv->sline = strdup(csv->line); /* copy original string */
  v = (csv->maxfield*sizeof(char *)); /* max size of field array */
  if ((csv->field = (char **)malloc(v)) == NULL)
    { /* problems */
      ErrMsg = "Can't allocate memory for array of fields";
      return WRPCSV_MEMORY;
    }
  for (i=0; i < csv->maxfield; i++) csv->field[i] = NULL;
  csv->numfields = 0;		/* no fields yet */

  p = csv->sline;		/* start of string to parse */

  do
    { /* parse string until done */
      if (csv->numfields >= csv->maxfield)
	{ /* too many fields */
	  ErrMsg = "Too many fields in string to parse";
	  return WRPCSV_INVALID;
	}

   while (isspace(*p)) p++; /* find start of field */
      if (*p == '"')
	{ /* field is quoted - find end of field */
	  q = advquoted(csv, ++p); /* skip initial quote */
	}
      else
	{ /* field not quoted, find next delimeter */
	  q = p + strcspn(p, csv->fieldsep); /* next delimeter */
	}

      /* end of field found - process this field */
      c = *q;			/* actual delimeter */
      *q = '\0';		/* terminate field */
      csv->field[csv->numfields++] = p; /* point to start of field */
      p = q;
      if (c != '\0') p++;	/* start of next field */
    } while (*p != '\0');	/* process rest of string */

  return 0;			/* just return zero for now */
}

/**********************************************************
 * Public Routines
 *********************************************************/

/*
 * Routines to affect global values
 */

char *wrpCsvGetErrMsg(void) { return ErrMsg; }

int wrpCsvCreate(void)
{ /* create new CSV ADT */
  int ret = WRPCSV_OK; /* so far, so good */

  /* if first time, initialize stuff */
  if (FirstTime)
    {
      if (init())
	{
	  ErrMsg = "Can't initialize CSV";
	  ret = WRPCSV_MEMORY;
	}
    }
  if (ret >= 0) ret = find_free_csv();	/* get next free entry */

  if (ret >= 0) /* negative value is error return */
    {
      Csv[ret] = create_csv();	/* create new csv struct */
      ret = Index[ret];	/* return index value instead of offset */
    }
  else
    {
      ErrMsg = "No free CSV cells";
    }

  return ret;
}

/* getter routines for global values */
int wrpCsvGetDefaultMaxLine(void) { return maxline; }
int wrpCsvGetDefaultMaxFields(void) { return maxfield; }
int wrpCsvGetMaxEnt(void) { return MaxEnt; }
char *wrpCsvGetDefaultDelims(void) { return fieldsep; }

/* setter routines for global values */
int wrpCsvSetDefaultMaxLine(int val)
{
  char *p;
  size_t v = val;		/* new maxline size */

  if (val <= 0) return WRPCSV_INVALID; /* must be positive */

  /* maxline must be the largest size of all CSVs */
  if ((p = realloc(InLine, v)) == NULL)
    { /* realloc failed */
      ErrMsg = "Not enough memory to re-allocate new input line";
      return WRPCSV_MEMORY; /* maxline is unchanged */
    }
  else
    { /* InLine re-allocated */
      InLine = p;		/* possibly new InLine */
    }
  maxline = val;		/* new max size */

  return WRPCSV_OK;
}

int wrpCsvSetDefaultMaxFields(int val)
{
  if (val <= 0) return WRPCSV_INVALID; /* must be positive  */

  maxfield = val;

  return WRPCSV_OK;
}

int wrpCsvSetDefaultDelims(char *delim)
{ /* change default delimeters */
  char *p;

  if ((p = strdup(delim)) == NULL)
    { /* out of memory */
      ErrMsg = "Can't allocate space for default delimeters";
      return WRPCSV_MEMORY;
    }
  fieldsep = p;			/* new list of delimeters */

  return WRPCSV_OK;
}

int wrpCsvSetMaxEnt(int val)
{
  struct wrpcsv **csv;
  size_t v;

  if (val <= 0) return WRPCSV_INVALID; /* must be positive */
  if ((csv = (struct wrpcsv **)realloc(Csv, v)) == NULL)
    { /* reallocation failed */
      ErrMsg = "Can't allocate new array for CSV structures";
      return WRPCSV_MEMORY;
    }
  /* space reallocated fine */
  Csv = csv;			/* new array */
  MaxEnt = val;			/* new max entries */

  return WRPCSV_OK;
}

/*
 * Routines to affect a specific CSV
 */

int wrpCsvDestroy(int key)
{
  struct wrpcsv *csv;
  int id;

  id = key % MaxEnt;
  if ((csv = get_csv(key)) == NULL) return WRPCSV_INVALID;

  Csv[id] = NULL;	/* remove csv from array */
  Index[id] = -1; /* free up index */
  free_csv(csv);	/* free up memory */

  return 0;
}

int wrpCsvReadLine(int id, FILE *fp)
{ /* read & parse line from file */
  int ret;
  struct wrpcsv *csv;

  /* validate parameters */
  if (fp == NULL) return WRPCSV_INVALID; /* bad FILE pointer */
  if ((csv = get_csv(id)) == NULL) return WRPCSV_INVALID; /* bad ID */

  /* read line, and then parse, using internal routine */
  if ((ret = readline(InLine, fp)) != WRPCSV_OK)
    {
      ErrMsg = "EOF on Input File";
      return WRPCSV_EOF;
    }

  if (csv->line != NULL) free(csv->line); /* free previous line */
  csv->line = strdup(InLine);	/* make copy of actual line */
  if (csv->line == NULL) 
    {
      ErrMsg = "Unable to allocate memory to copy string";
      return WRPCSV_MEMORY; /* out of memory */
    }

  ret = parse_csv(csv);		/* parse the line */

  return ret;
}

int wrpCsvParseString(int id, char *p)
{ /* parse the given string */
  int ret = WRPCSV_OK;
  struct wrpcsv *csv;

  if (p == NULL)
    {
      ErrMsg = "NULL String Pointer passed to ParseString";
      return WRPCSV_INVALID;	/* bad string */
    }

  if ((csv = get_csv(id)) == NULL) return WRPCSV_INVALID; /* bad handle */

  if (csv->line) free(csv->line); /* free previous line */
  csv->line = strdup(p);	/* make copy of string */
  ret = parse_csv(csv);		/* parse the string */

  return ret;
}

char *wrpCsvGetLine(int id)
{ /* return the original line as a string */
  struct wrpcsv *csv;

  if ((csv = get_csv(id)) == NULL) return NULL;
  
  return csv->line;
}

int wrpCsvGetNumFields(int id)
{
  struct wrpcsv *csv;

  if ((csv = get_csv(id)) == NULL) return WRPCSV_INVALID;

  return csv->numfields;
}

char *wrpCsvGetField(int id, int ndx)
{
  struct wrpcsv *csv;

  if ((csv = get_csv(id)) == NULL) return NULL;
  if (ndx < 0 || ndx >= csv->numfields) return NULL;

  return csv->field[ndx];
}

int wrpCsvSetMaxLine(int id, int val)
{
  struct wrpcsv *csv;

  if (val <= 0)
    {
      ErrMsg = "Invalid MaxLine value";
      return WRPCSV_INVALID;
    }
  if ((csv = get_csv(id)) == NULL) return WRPCSV_INVALID;

  if (val > maxline)
    { /* new value is bigger - allocate more space */
      if (wrpCsvSetDefaultMaxLine(val) == WRPCSV_OK)
	{ /* max line changed, all is well */
	}
      else
	{ /* max line not changed - return error */
	  ErrMsg = "Unable to change default MaxLine";
	  return WRPCSV_INVALID;
	}
    }
  csv->maxline = val;

  return WRPCSV_OK;
}
